package at.ac.uibk.gitsearch.repository.search;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PostConstruct;

import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.MultiMatchQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightBuilder;
import org.springframework.stereotype.Repository;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import at.ac.uibk.gitsearch.config.ApplicationProperties;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.GitProject;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.Person;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.SearchResultDTO;

@Repository
public class MetaDataRepository {
	
	/**
	 * just a local wrapper, to decrypt JSON
	 *
	 */
	public static class SearchResultDTOWrapper {
		private SearchResultDTO metadata = new SearchResultDTO();

		public SearchResultDTO getMetadata() {
			return metadata;
		}

		public void setMetadata(SearchResultDTO metadata) {
			this.metadata = metadata;
		}

		private GitProject project;

		public GitProject getProject() {
			return project;
		}

		public void setProject(GitProject project) {
			this.project = project;
		}

	}

	private final RestHighLevelClient elasticsearchClient;

	private final ApplicationProperties properties;

	private static final int FRAGMENT_SIZE = 5000;

	public MetaDataRepository(RestHighLevelClient elasticsearchClient, ApplicationProperties properties) {
		this.elasticsearchClient = elasticsearchClient;
		this.properties = properties;

	}

	/**
	 * contains for each Field, a list of Strings mapped to a completion. Together
	 * with hit counts E.g. "metadata.creator" -> Kerstin -> {"Kerstin Limbeck" ->
	 * 3, "Kerstin Mair" -> 2)}
	 */
	private Map<String, Map<String, Map<String, Integer>>> cachedCompletions = null;

	private static final int MAX_AUTO_COMPLETION_RESULTS = 10;

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix) throws IOException {
		return getFieldAutoCompletion(keyWordPrefix, SearchRepositoryConstants.METADATA_KEYWORDS);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param progammingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String progammingLanguagePrefix) throws IOException {
		return getFieldAutoCompletion(progammingLanguagePrefix,
				SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix) throws IOException {
		return getFieldAutoCompletion(creatorPrefix, SearchRepositoryConstants.METADATA_CREATOR);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix) throws IOException {
		return getFieldAutoCompletion(contributorPrefix, SearchRepositoryConstants.METADATA_CONTRIBUTOR);
	}

	/**
	 * returns a list of hits together with the number of occurences
	 * @param keyWordPrefix
	 * @param metaDataField
	 * @return
	 * @throws IOException
	 * @throws JsonProcessingException
	 * @throws JsonMappingException
	 */
	private List<AutoCompleteEntry> getFieldAutoCompletion(String keyWordPrefix, final String metaDataField)
			throws IOException, JsonProcessingException, JsonMappingException {
		fillAutoCompletion();

		// höllisch kompliziert und höllisch unverständlich mit Streams :-)
		// idea
		// 1. Filtern nach Prefix
		// 2. Zusammenzählen der möglichen Hits (Besser Max statt Sum?)
		// 3. Auf MAX_AUTO_COMPLETION_RESULTS begrenzen und nur die Trefferstrings
		// ausgeben.
		final Stream<Entry<String, Map<String, Integer>>> filteredEntries = cachedCompletions.get(metaDataField)
				.entrySet().stream() // Map<String, Map<String, Integer>> -> Stream<Entry<String, Map<String,
										// Integer>>
				.filter(s -> s.getKey().startsWith(keyWordPrefix));
		final Map<String, Integer> combinedHits = filteredEntries // s = String -> bool -> Stream<Entry<String,
																	// Map<String, Integer>>
				.flatMap(e -> e.getValue().entrySet().stream()) // Stream< Map<String, Integer>> with duplicates
				.collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue, Integer::max));
		return combinedHits.entrySet().stream().sorted((e1, e2) -> e2.getValue() - e1.getValue())
				.limit(MAX_AUTO_COMPLETION_RESULTS).map(s -> new AutoCompleteEntry(s.getKey(), s.getValue())).collect(Collectors.toList()); // Map.Entry<String,
																										// Map<String,
																										// Integer>>

	}

	@PostConstruct
	private synchronized void fillAutoCompletion() throws IOException, JsonProcessingException, JsonMappingException {
		if (cachedCompletions == null) {
			Map<String, Map<String, Map<String, Integer>>> reloadedCachedCompletions = new HashMap<>();
			reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_KEYWORDS, new TreeMap<>());
			reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CREATOR, new TreeMap<>());
			reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CONTRIBUTOR, new TreeMap<>());
			reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, new TreeMap<>());

			SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

			final int pageSize = 20;
			int currentPage = 0;
			boolean tryNextPage = true;

			while (tryNextPage) {

				SearchSourceBuilder sourceBuilder = new SearchSourceBuilder().size(pageSize).from(currentPage++*pageSize);
				
				searchRequest.source(sourceBuilder);

				SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

				ObjectMapper objectMapper = new ObjectMapper();
				objectMapper.registerModule(new JavaTimeModule());
				objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

				tryNextPage = searchResponse.getHits().getHits().length > 0;
				
				for (SearchHit searchHit : searchResponse.getHits().getHits()) {
					final String sourceAsJSON = searchHit.getSourceAsString();
					SearchResultDTOWrapper entry = objectMapper.readValue(sourceAsJSON, SearchResultDTOWrapper.class);
					if (entry.getMetadata().getKeyword() != null)
						for (String keyWord : entry.getMetadata().getKeyword())
							addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_KEYWORDS),
									split(keyWord), keyWord);
					if (entry.getMetadata().getContributor() != null)
						for (Person contributor : entry.getMetadata().getContributor())
							addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CONTRIBUTOR),
									split(contributor.getName()), contributor.getName());
					if (entry.getMetadata().getCreator() != null)
						for (Person creator : entry.getMetadata().getCreator())
							addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CREATOR),
									split(creator.getName()), creator.getName());
					if (entry.getMetadata().getProgrammingLanguage() != null)
						for (String language : entry.getMetadata().getProgrammingLanguage())
							addTo(reloadedCachedCompletions
									.get(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES),
									Collections.singletonList(language), language);
				}
			}
			cachedCompletions = reloadedCachedCompletions;
		}
	}

	private void addTo(Map<String, Map<String, Integer>> completionMap, List<String> tokenList, String target) {
		tokenList.forEach(token -> {
			Map<String, Integer> existingTargets = completionMap.get(token);
			if (existingTargets == null) { // none found
				existingTargets = new HashMap<>();
				existingTargets.put(target, 1);
				completionMap.put(token, existingTargets);
			} else {
				final Integer count = existingTargets.get(target);
				if (count == null) {
					existingTargets.put(target, 1);
				} else {
					existingTargets.put(target, count + 1);
				}

			}
//			existingTargets.add(target);
		});
	}

	private static List<String> split(String s) {
		StringTokenizer st = new StringTokenizer(s, " \t\n\r\f ");
		List<String> result = new ArrayList<>();
		while (st.hasMoreTokens()) {
			result.add(st.nextToken());
		}
		return result;
	}

	public SearchResultsDTO pageDetails(SearchInputDTO searchInputDTO, int pageSize) throws IOException {
		fillAutoCompletion();
		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getFulltextQuery(),
				term -> queryBuilder.must(QueryBuilders
						.multiMatchQuery(term, SearchRepositoryConstants.METADATA_DESCRIPTION,
								SearchRepositoryConstants.METADATA_KEYWORDS, SearchRepositoryConstants.METADATA_TITLE)
						.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX)));

		forEachTerm(searchInputDTO.getMetadata().getProgrammingLanguage(), term -> queryBuilder
				.must(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, term)));

		forEachTerm(searchInputDTO.getMetadata().getKeyword(), term -> queryBuilder
				.must(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_KEYWORDS, term)));

		if (searchInputDTO.getMetadata().getAuthor() != null)
			queryBuilder.must(QueryBuilders
					.multiMatchQuery(searchInputDTO.getMetadata().getAuthor(),
							SearchRepositoryConstants.METADATA_CREATOR, SearchRepositoryConstants.METADATA_CONTRIBUTOR,
							SearchRepositoryConstants.METADATA_PUBLISHER)
					.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX));

//		forEachTerm(searchInputDTO.getMetadata().getAuthor(),		
//		   term ->
// 			queryBuilder.must(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_CREATOR, term)));

		forEachTerm(searchInputDTO.getMetadata().getLicense(),
				term -> queryBuilder.must(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_LICENSE, term)));

		char[] boundaryChars = { '\n' };

		HighlightBuilder highlightBuilder = new HighlightBuilder().highlighterType("plain")
				.preTags(properties.getSearch().getHighlightPre()).postTags(properties.getSearch().getHighlightPost())
				.field(SearchRepositoryConstants.METADATA_DESCRIPTION).boundaryChars(boundaryChars)
				.boundaryScannerType(HighlightBuilder.BoundaryScannerType.SENTENCE).fragmentSize(FRAGMENT_SIZE);

		// String[] includes = {};
		// String[] excludes = {HighlightRepositoryConstants.CONTENT_FIELD};

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder).highlighter(highlightBuilder).size(pageSize)
				.from((searchInputDTO.getPage()) * pageSize);
		// .fetchSource(includes, excludes);

		searchRequest.source(sourceBuilder);

		SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

		ObjectMapper objectMapper = new ObjectMapper();
		objectMapper.registerModule(new JavaTimeModule());
		objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

		final SearchResultsDTO results = new SearchResultsDTO();
		List<SearchResultDTO> searchResults = new ArrayList<>();
		for (SearchHit searchHit : searchResponse.getHits().getHits()) {
			SearchResultDTOWrapper entry = objectMapper.readValue(searchHit.getSourceAsString(),
					SearchResultDTOWrapper.class);
			final SearchResultDTO metadata = entry.getMetadata();
			metadata.setProject(entry.getProject());
			searchResults.add(metadata);
		}
		long hitCount = searchResponse.getHits().getTotalHits();
		results.setHitCount(hitCount);
		results.setSearchResult(searchResults);
		return results;
	}

	/** Helper **/
	private void forEachTerm(String searchTerm, Consumer<String> exec) {
		if (searchTerm == null)
			return;
		StringTokenizer st = new StringTokenizer(searchTerm);
		while (st.hasMoreTokens()) {
			exec.accept(st.nextToken());
		}
	}

}
